; Rheinturm-Uhr Version 2.0.1

; Quartz 12 Mhz
; Mikrocontroller AT89C1051 oder AT89C2051
; Die Led's werden als 5x9 Matrix angesteuert.
; Die Tasten werden als 5x1 Matrix angesteuert.

; Timer 0 erzeugt den Sekundentakt, steuert die Matrix und berechnet
; den bertrag der Sekunden-Einer auf hhere Stellen.
; Das Hauptprogramm empfngt das DCF Signal.

; Das Hauptprogramm versucht, wenigstens 2 mal hintereinander die gleiche
; Uhrzeit zu empfangen. Wenn das gelingt, wird die Uhr entsprechend
; eingestellt. Der DCF Empfang wird staendig wiederholt.
; Die Trenn-Led's zeigen das DCF Signal an, wenn die Uhr nicht Synchron
; luft.

; Matrix Spalten

.equ	Led1,0x97	; P1.7    Led's 1-9
.equ	Led2,0x96	; P1.6
.equ	Led3,0x95	; P1.5
.equ	Led4,0x94	; P1.4
.equ	Led5,0x93	; P1.3
.equ	Led6,0x92	; P1.2
.equ	Led7,0x91	; P1.1
.equ	Led8,0x90	; P1.0
.equ	Led9,0xB7	; P3.7

.equ	Led10,0x92	; P1.2	 Led's 10 und 20 Stunden
.equ	Led20,0x93	; P1.3

.equ    LedRot,0xB7	; P3.7  Trenn-Led's

; Matrix Reihen

.equ	SekundenE,0xB5	; P3.5	Sekunden Einer
.equ	SekundenZ,0xB3	; P3.3	Sekunden Zehner
.equ	MinutenE,0xB2	; P3.2  Minuten Einer
.equ	MinutenZ,0xB1	; P3.1  Minuten Zehner
.equ	StundenE,0xB4	; P3.4  Stunden Einer
.equ	StundenZ,0xB1	; P3.1  Stunden Zehner = Minuten Zehner !

; Matrix der Tasten

.equ	Tasten,0xB0	; P3.0  Einstelltasten
.equ    Taste1,0xB5	; P3.5  Sekunden 00
.equ	Taste2,0xB2	; P3.2  Minuten Einer
.equ	Taste3,0xB3	; P3.3  Minuten Zehner
.equ	Taste4,0xB4	; P3.4  Stunden Einer
.equ	Taste5,0xB1	; P3.1  Stunden Zehner

; Sonstige

.equ	DCF,0xB0		; P3.0 , DCF77 Signal, mit Tastenmatrix geteilt

; fest verwendete Bit-Adressen

.equ    Taste_gedrueckt,0xBB	; High, wenn gerade eine Taste gedrckt wird
.equ	Synchron,0xBA		; High, wenn die Uhr DCF Synchrom luft

; Fest verwendete Register in Registersatz 0 (Interrupt)

; R7	Interrupt-Zaehler fr Sekundentakt
; R6	Stunden Zehner
; R5	Stunden Einer
; R4	Minuten Zehner
; R3	Minuten Einer
; R2	Sekunden Zehner
; R1	Sekunden Einer
; R0	Interruptzaehler fr Matrix 

; Fest verwendete Register in Registersatz 1 (Hauptprogramm)

; R7	Stunden vorheriger Empfang
; R6	Minuten vorheriger Empfang
; R5	Stunden aktuell
; R4	Minuten aktuell
; R3	Bitnummer
; R0	Timer fr DCF Impulsbreitenmessung. Wird mit 125 Hz hochgezaehlt.


.org 0x0000		; Reset 
	sjmp start


.org 0x000B		; Timer 0 Interrupt
	push PSW	; Flags sichern
	clr  RS0
	mov  TH0,#206	; Timer auf 32*50 stellen -> 625 Hz
	acall matrix
	djnz R0,ti0	; Takt durch 5 Teilen -> 125 Hz
	mov  R0,#5
	inc  8		; Timer fr DCF-Messungen hochzaehlen (R0 in RS1)
	djnz R7,ti0	; Takt durch 125 Teilen -> 1 Hz
	mov  R7,#125	
	inc  R1		; Sekunden +1
	acall korrektur
ti0:	pop  PSW
	reti


start:
	mov  SP,#16	; SP hochsetzen, weil 2 Registersaetze gebraucht werden	mov  
TMOD,#0	; Beide Timer Modus 0, 5+8 Bit
	mov  TMOD,#0	; Timer Modus 0 (5+8 Bit)
	setb EA		; Timer Interrupts einschalten
	setb ET0
	clr  A		; Uhr auf 0 stellen
	mov  R1,A
	mov  R2,A
	mov  R3,A
	mov  R4,A
	mov  R5,A
	mov  R6,A
	setb TR0	; Timer starten
	setb RS0	; Verwende Registersatz 1 im Hauptprogramm

hauptprogramm:		  ; DCF Signal empfangen
	mov   R7,13	  ; 13=R5. Vorherige Uhrzeit sichern
	mov   R6,12	  ; 12=R4
	acall WarteStart
	mov   R3,#0
hp1:	acall EmpfangeBit
	cjne  R3,#21,hp1
	mov   R3,#0
hp2:	acall EmpfangeBit ; Minuten empfangen
	rrc   A
	cjne  R3,#8,hp2
	clr   ACC.7	  ; Das Prfbit interessiert uns nicht
	mov   R4,A	  ; Minuten speichern
	mov   R3,#0
hp3:	acall EmpfangeBit ; Stunden empfangen
	rrc   A
	cjne  R3,#8,hp3	
	clr   ACC.6	; Das Prfbit interessiert uns nicht
	clr   ACC.7	; Das Bit 0 vom Kalendertag interessiert uns auch nicht
	mov   R5,A	; Stunden speichern

			; jetzt prfen, ob zweimal die selbe Uhrzeit empfangen wurde
	mov   A,R6
	add   A,#1	; dabei beruecksichtigen, dass die Minuten um 1 abweichen
	da    A
	cjne  A,#96,hp4 
	mov   A,#0
hp4:	clr   C
	subb  A,R4
	jnz   Nicht_Synchron
	mov   A,R7
	clr   C
	subb  A,R5
	jnz   Nicht_Synchron

Ist_Synchron:		; Uhrzeit wurde 2x erfolgreich empfangen. Uhr einstellen
	setb Synchron
	clr  RS0	; Sekunden einstellen (auf 37)
	mov  R1,#7
	mov  R2,#3
	mov  R7,#0
	setb RS0
	mov  A,R6	; Minuten einstellen
	anl  A,#15
	mov  3,A	; 3=R3 in Registersatz 0
	mov  A,R4
	swap A
	anl  A,#7
	mov  4,A	; 4=R4 in Registersatz 0
	mov  A,R7	; Stunden einstellen
	anl  A,#15
	mov  5,A	; 5=R5 in Registersatz 0
	mov  A,R5
	swap A
	anl  A,#3
	mov  6,A	; 6=R6 in Registersatz 0
	ajmp hauptprogramm

Nicht_Synchron:		; Empfangsfehler
	clr Synchron
	ajmp hauptprogramm


WarteStart:		; Warte auf den fehlenden Impuls (59. Sekunde)
	mov  R0,#0	
ws1:	jnb  DCF,wartestart
	cjne R0,#187,ws1 ; War die Impulspause>1500ms?
	ret


EmpfangeBit:		; Empfngt ein Bit. Liefert den Wert in C
	push ACC
	mov  R0,#0
eb2:	mov  A,R0	; War die Pause zu lang (1100ms)?
	clr  C
	subb A,#140
	jc   eb3
	clr  Synchron   ; Zu lange gewartet, Empfangsfehler. Uhr ist nicht mehr 
Synchron
	mov  R7,#255	; vorherige Stunden=255. Dieser ungueltige Wert verhindert,
			; dass die Uhr trotz Empfangsfehler eingestellt wird.
	ret
eb3:	jb   DCF,eb2
	mov  A,R0	; War die Pause lang genug (mindestens 50ms)?
	clr  C
	subb A,#6
	jc   eb2        ; zu Kurz, nochmal
	inc  R3
	mov  R0,#0
eb1:	jnb  DCF,eb1
	mov  A,R0	; War der Impuls lang genug (mindestens 50ms)?
	clr  C
	subb A,#6
	jc   eb1	; Zu kurz, weiter warten
	mov  A,R0
	clr  C
	subb A,#16
	cpl  C
	pop  ACC
	ret
	
matrix:			; Matrix anzeigen
	push ACC
	setb SekundenE	; Alle Led's aus schalten
	setb SekundenZ
	setb MinutenE
	setb MinutenZ
	setb StundenE
	mov  A,R0	; Schritt-Nummer ermitteln (0-5)

ma0:	cjne A,#1,ma1
	mov  A,R1	; Sekunden Einer
	acall dezimal
	clr  SekundenE  
	ajmp matrixende

ma1:	cjne A,#2,ma2
	acall tasten_abfrage
	mov  A,R2	; Sekunden Zehner
	acall dezimal
	mov  C,DCF 	; und untere Trenn-Led
	orl  C,Synchron
	cpl  C
	mov  LedRot,C
	clr  SekundenZ
	ajmp matrixende

ma2:	cjne A,#3,ma3
	mov  A,R3	; Minuten Einer
	acall dezimal
	clr  MinutenE
	ajmp matrixende

ma3:	cjne A,#4,ma4
	mov  A,R4	; Minuten Zehner
	acall dezimal
	mov  A,R6	; und Stunden Zehner
	acall dezimal2
	mov  C,DCF	; und obere Trenn-Led
	orl  C,Synchron
	cpl  C
	mov  LedRot,C
	clr  MinutenZ
	ajmp matrixende

ma4:	mov  A,R5	; Stunden Einer
	acall dezimal
	clr  StundenE
	ajmp matrixende

matrixende:
	pop ACC
	ret

tasten_abfrage:
	setb Led1	; Tasten abfragen
	setb Led2
	setb Led3
	setb Led4
	setb Led5
	setb Led6
	setb Led7
	setb Led8
	setb Led9
	clr  Tasten
	jb   Taste1,ta1
	mov  R1,#0
	mov  R2,#0
	sjmp te2
ta1:	jb   Taste2,ta2
	jb   Taste_gedrueckt,te1
	inc  R3
	sjmp te2
ta2:	jb   Taste3,ta3
	jb   Taste_gedrueckt,te1
	inc  R4
	sjmp te2
ta3:	jb   Taste4,ta4
	jb   Taste_gedrueckt,te1
	inc  R5
	sjmp te2
ta4:	jb   Taste5,ta5
	jb   Taste_gedrueckt,te1
	inc  R6
	sjmp te2
ta5:	setb Tasten
	clr  Taste_gedrueckt		; Es wurde keine Taste gedrckt
	ret	
te2:	setb Taste_gedrueckt		; Es wurde eine Taste gedrckt
	acall korrektur
te1:	setb Tasten			; Es wird auf Loslassen gewartet
	ret

korrektur:		; Laesst die Zhler berlaufen. Mit bertrag
	cjne R1,#10,s1
	mov  R1,#0
	inc  R2		; Sekunden +10
s1:	cjne R2,#6,s2
	mov  R2,#0
	inc  R3		; Minuten +1
s2:	cjne R3,#10,s3
	mov  R3,#0
	inc  R4		; Minuten +10
s3:	cjne R4,#6,s4
	mov  R4,#0
	inc  R5		; Stunden +1
s4:	cjne R5,#10,s5
	mov  R5,#0
	inc  R6		; Stunden +10
s5:	cjne R6,#3,s6
	mov  R6,#0
s6:	cjne R5,#4,s7	; auerdem bei 24 Stunden auf 0 gehen
	cjne R6,#2,s7
	mov  R5,#0
	mov  R6,#0
s7:	ret


dezimal:		; Gibt A als Dezimalzahl auf Led's aus
	setb Led1
	setb Led2
	setb Led3
	setb Led4
	setb Led5
	setb Led6
	setb Led7
	setb Led8
	setb Led9
        jz   dz1
	clr  Led1
	dec  A
	jz   dz1
	clr  Led2
	dec  A
	jz   dz1
	clr  Led3
	dec  A
	jz   dz1
	clr  Led4
	dec  A
	jz   dz1
	clr  Led5
	dec  A
	jz   dz1
	clr  Led6
	dec  A
	jz   dz1
	clr  Led7
	dec  A
	jz   dz1
	clr  Led8
	dec  A
	jz   dz1
	clr  Led9
dz1:	ret


dezimal2:		; Gibt A als Dezimalzahl fr Stunden Zehner aus
        jz   dz2
	clr  Led6
	dec  A
	jz   dz2
	clr  Led7
dz2:	ret


